/* This stylesheet is used for printing only, overriding article-style.css */

html,
body,
.gdarticle {
  background: white;
}

/* Hide audio icons */

.dsl_s_wav {
  display: none;
}

/* hide collapse area*/

.collapse_expand_area {
  display: none;
}
