html {
  background-color: white;
}

body {
  background: white;
}

a:hover {
  text-decoration: underline;
}

.gdarticle {
  background: white;
}

/* The 'From ' string which preceeds dictionary name in the heading */
.gdfromprefix {
  display: none;
}

/* Dictionary's name heading */
.gddictname {
  padding: 0.2em;
  padding-left: 0.5em;
  margin-bottom: 0.1em;
  font-size: 14px;
  font-weight: normal;
  float: right;
  border: 1px solid white;
  margin-top: 7px;
  color: #7f7f7f;
  background: white;
}

.gdarticleseparator + script + .gdarticle .gddictname {
  border-top: 1px solid #ccc;
  /*  margin-top: 11px;*/
}

.gdactivearticle .gddictname {
  font-weight: normal;

  float: right;
  border: 1px solid #b3d7f7;
  margin-top: 7px;
  color: #0066ff;
  background: #deecf8;
}

.gdarticleseparator + script + .gdactivearticle .gddictname {
  border-top: 1px solid #b3d7f7;
  /*  margin-top: 11px;*/
}

.gdarticleseparator {
  margin-top: 10px;
  border-bottom: 1px solid #cccccc;
  margin-bottom: 0px;
  width: 100%;
  display: inline-block;
  clear: both;
}

.gddictname + h3,
.gddictname + .dsl_article > .dsl_headwords {
  margin-top: 10px;
}

.dsl_ex,
.dsl_ex .dsl_opt {
  color: inherit;
}

.dsl_com {
  color: black;
}
