/* This file is (c) 2010 Jennie Petoumenou <epetoumenou@gmail.com>
 * Part of GoldenDict. Licensed under GPLv3 or later, see the LICENSE file */

#pragma once

#include "dictionary.hh"

// Support for Greek transliteration
namespace GreekTranslit {

sptr< Dictionary::Class > makeDictionary();
}
