/* remove the main toolbar handle */
#navToolbar::handle,
#dictionaryBar::handle {
  image: none;
}

#navToolbar #scanPopupButton {
  margin-left: 5px;
}

/* Colored header for the "Found in Dictionaries:" pane */
#dictsPaneTitleBar {
  margin: 2px;
}

/* Colored header for the History Pane */
#historyPaneTitleBar {
  margin: 2px;
}

/* Thin separators between the main UI and sidebars */
QMainWindow::separator {
  width: 1px; /* when vertical */
  height: 1px; /* when horizontal */
}
